#include "starters.h"
#include "SDL.h"
#pragma once
class GeomObj;
//@@@@@@@@@@@@@@@@@@@ PointCluster class @@@@@@@@@@@@@@@@@@
class PointCluster { // holds array of points for the bounding hull of a shape
public:
	int num;
	Point3* pt;
	PointCluster() { num = 0; pt = NULL; }
	PointCluster(int n) { // make a cluster of n points
		pt = new Point3[n]; assert(pt);
		num = n;
	}
};

//@@@@@@@@@@@@@@@@@@@@@@@@@@ SphereInfo @@@@@@@@@@@@@@@@@@@@@
class SphereInfo { // holds the center and radius of a sphere
public:
	Point3 center;
	float radSq;
	void set(float x, float y, float z, float rsq) {
		center.set(x,y,z);
		radSq = rsq;
	}
};

//@@@@@@@@@@@@@@@@@@@@@@@@ Cuboid @@@@@@@@@@@@@@@@@@@@@@@@
class Cuboid { // holds six border values of a cuboid
public:
	float left, top, right, bott, front, back;
	void set(float l, float t, float r, float b, float f, float bk) {
		left = l; top = t; right = r; bott = b; front = f; back = bk;
	}
	void set(Cuboid& c) {
		left = c.left; top = c.top; right = c.right; bott = c.bott;
		front = c.front; back = c.back;
	}
	void print() {
		cout << "\n(l,t,r,b,f,bk) = (" << left << "," << top << "," << right 
			 << "," << bott << "," << front << "," << back << ")";
	}
};

//@@@@@@@@@@@@@@@@@@@@ Ray @@@@@@@@@@@@@@@@@@@@@@@@@
class Ray {
public:
	Point3 start;
	Vector3 dir;
	int recurseLevel;
	int row, col; // for screen extents
	int numInside; // number of objects on list
	GeomObj* inside[10]; // array of object pointers

	Ray() {	// default constructor
		start.set(0,0,0);
		dir.set(0,0,0);
		numInside = 0;
	}
	Ray(Point3 origin) { // constructor: set start point of ray
		start.set(origin);
		dir.set(0,0,0);
		numInside = 0;
	}
	Ray(Point3& origin, Vector3& direction) { // constructor: set start and dir of ray
		start.set(origin); 
		dir.set(direction);
		numInside = 0;
	}
	Point3 rayPos(double time) {
		return Point3(start.x+time*dir.x, start.y+time*dir.y, start.z+time*dir.z);
	}
	Ray xfrmRay(Affine4 T) {
		return Ray(Point3(start.x*T.m[0]+start.y*T.m[1]+start.z*T.m[2],
						start.x*T.m[4]+start.y*T.m[5]+start.z*T.m[6],
						start.x*T.m[8]+start.y*T.m[9]+start.z*T.m[10]),
						Vector3(dir.x*T.m[0]+dir.y*T.m[1]+dir.z*T.m[2],
						dir.x*T.m[4]+dir.y*T.m[5]+dir.z*T.m[6],
						dir.x*T.m[8]+dir.y*T.m[9]+dir.z*T.m[10]));
	}
	void setStart(Point3& p) { start.set(p); }
	void setDir(float x, float y, float z) { dir.x = x; dir.y = y; dir.z = z; }
	void setRayDirection(Light *L); // for shadow feelers
	void setRayDirection(Vector3& dir); // for spawned rays
	int isInShadow();
	void makeGenericRay(GeomObj* p, Ray& gr);
};

//@@@@@@@@@@@@@@@@@@ HitInfo @@@@@@@@@@@@@@@@@@@@
class HitInfo { // data for each hit with a surface
  public:
	double hitTime;    		// the hit time
	GeomObj* hitObject;  	// the object hit
	int surface;			    // which surface is hit?
	int isEntering;			 // is ray entering the object?
	Point3 hitPoint;		   // hit point 
	Vector3 hitNormal;		// normal at hit point
	HitInfo() 	{ 
		hitObject = NULL;
		hitTime = -1000;
		surface = 0;
		isEntering = 0;
	}
	void set(HitInfo& h) {
		hitTime = h.hitTime; hitObject = h.hitObject;	
		surface = h.surface; isEntering = h.isEntering;
		hitPoint.set(h.hitPoint); hitNormal.set(h.hitNormal);
	}
};

//@@@@@@@@@@@@@@@@@ Intersection @@@@@@@@@@@@@@@@@@@@@
class Intersection { // hold the hit list 
 public:
	#define maxNumHits 8
	int numHits;			 // the number of hits
	HitInfo hit[maxNumHits];	 // list of hits;
	Intersection() { numHits = 0; }  // default constructor
	void set(Intersection& intr) { // copy intersection info
		numHits = intr.numHits;
		for(int i = 0; i < maxNumHits; i++)
			hit[i].set(intr.hit[i]);
	}
};
