#include "SDL.h"

//<<<<<<<<<<<<<<<<<< hit() for Cube >>>>>>>>>>>>>>>>>>>>>>>
bool Cube::hit(Ray &r, Intersection &inter)  {
	double tHit, numer, denom;
	double tIn = -100000.0, tOut = 100000.0;	// plus-minus infinity?
	Ray genRay;
	int inSurf, outSurf;	// which of the six surfaces
	genRay = r.xfrmRay(invTransf);
	for (int i=0; i<6; i++) {
		switch (i) {	// which plane of cube to test
			case 0:
				numer = 1.0 - genRay.start.y;
				denom = genRay.dir.y;
				break;
			case 1:
				numer = 1.0 + genRay.start.y;
				denom = -genRay.dir.y;
				break;
			case 2:
				numer = 1.0 - genRay.start.x;
				denom = genRay.dir.x;
				break;
			case 3:
				numer = 1.0 + genRay.start.x;
				denom = -genRay.dir.x;
				break;
			case 4:
				numer = 1.0 - genRay.start.z;
				denom = genRay.dir.z;
				break;
			case 5:
				numer = 1.0 + genRay.start.z;
				denom = -genRay.dir.z;
				break;
		}
		if (fabs(denom) < 0.00001) {	// ray is parallel
			if (numer < 0) return false;	// ray is out
		} else {	// ray is not parallel
			tHit = numer / denom;
			if (denom > 0) {	// exitting
				if (tHit < tOut) {	// a new earlier exit
					tOut = tHit;
					outSurf = i;
				}
			} else {	// denom is negative: entering
				if (tHit > tIn) {	// a later entrance
					tIn = tHit;
					inSurf = i;
				}
			}
		}
		if (tIn >= tOut) return false;	// it's a miss - early out
	}	// end of for loop
	int num = 0;	// no positive hits yet
	if (tIn > 0.00001) {	// is first hit in front of the eye?
		inter.hit[0].hitTime = tIn;
		inter.hit[0].surface = inSurf;
		inter.hit[0].isEntering = 1;
		inter.hit[0].hitObject = this;
		inter.hit[0].hitPoint.set(genRay.rayPos(tIn));
		inter.hit[0].hitNormal.set(cubeNormal(inSurf));
		num++;	// have a hit
	}
	if (tOut > 0.00001) {
		inter.hit[0].hitTime = tOut;
		inter.hit[0].surface = outSurf;
		inter.hit[0].isEntering = 0;
		inter.hit[0].hitObject = this;
		inter.hit[0].hitPoint.set(genRay.rayPos(tOut));
		inter.hit[0].hitNormal.set(cubeNormal(outSurf));
		num++;	// have a hit
	}
	inter.numHits = num;	// number of hits in front of the eye
	return (num > 0);
}

//<<<<<<<<<<<<<<<<<< hit() for UnionBool >>>>>>>>>>>>>>
bool UnionBool::hit(Ray &r, Intersection &inter) {
	Intersection lftInter, rtInter;
	if (!left->hit(r,lftInter) && !right->hit(r,rtInter)) { return false; }	// misses outer object
	GeomObj *leftObj = lftInter.hit[0].hitObject, *rightObj = rtInter.hit[0].hitObject;
	if (!right->hit(r,rtInter)) {	// only hits left obj
		inter = lftInter;
		return true;
	} else if (!left->hit(r,lftInter)) {	// only hits right obj
		inter = rtInter;
		return true;
	} else {
		bool lftInside = !(lftInter.hit[0].isEntering);	// initial insideness
		bool rtInside = !(rtInter.hit[0].isEntering);
		bool combInside = lftInside && rtInside;	// for the difference

		// go through the list until one has been consumed
		int iL=0, iR=0, iC=-1;
		HitInfo nextHit;
		while ((iL < lftInter.numHits) && (iR < rtInter.numHits)) {
			bool newInside;	// new result state
			if ((lftInter.hit[iL].hitTime <= rtInter.hit[iR].hitTime) && (iL < lftInter.numHits)) {
				// hits left first and still left obj to hit
				nextHit = lftInter.hit[iL++];
				lftInside = !lftInside;
				newInside = lftInside && rtInside;
 				leftObj = nextHit.hitObject;
				if (newInside != combInside) {
					inter.hit[++iC] = nextHit;
					combInside = newInside;
					inter.hit[iC].isEntering = combInside;
				}
			} else {	// hits right first, or run out of left to hit
				nextHit = rtInter.hit[iR++];
				rtInside = !rtInside;
				newInside = lftInside && rtInside;
				if (newInside != combInside) {
					inter.hit[++iC] = nextHit;
					combInside = newInside;
					inter.hit[iC].isEntering = combInside;
					inter.hit[iC].hitObject = rightObj;
					/*if (combInside) {
						nextHit.hitNormal.flip();
						inter.hit[iC].hitObject = leftObj;
					}*/
				}
			} // end ELSE
		} // end WHILE
	} // end ELSE
	return (inter.numHits > 0);
}
//<<<<<<<<<<<<<<<<<< hit() for DifferenceBool >>>>>>>>>>>>>>
bool DifferenceBool::hit(Ray &r, Intersection &inter) {
	Intersection lftInter, rtInter;
	lftInter.numHits = rtInter.numHits = 0;
	if (!left->hit(r,lftInter)) { return false; }	// misses outer object
	GeomObj* leftObj = lftInter.hit[0].hitObject;
	if (!right->hit(r,rtInter)) {
		inter = lftInter;
		return true;
	}
	bool lftInside = !(lftInter.hit[0].isEntering);	// initial insideness
	bool rtInside = !(rtInter.hit[0].isEntering);
	bool combInside = lftInside && !rtInside;	// for the difference

	// go through the list until one has been consumed
	int iL=0, iR=0, iC=-1;
	HitInfo nextHit;
	while ((iL < lftInter.numHits) && (iR < rtInter.numHits)) {
		bool newInside;	// new result state
		if (lftInter.hit[iL].hitTime <= rtInter.hit[iR].hitTime) {
			nextHit = lftInter.hit[iL++];
			lftInside = !lftInside;
			newInside = lftInside && !rtInside;
 			leftObj = nextHit.hitObject;
			if (newInside != combInside) {
				inter.hit[++iC] = nextHit;
				combInside = newInside;
				inter.hit[iC].isEntering = combInside;
			}
		} else {
			nextHit = rtInter.hit[iR++];
			rtInside = !rtInside;
			newInside = lftInside && !rtInside;
			if (newInside != combInside) {
				inter.hit[++iC] = nextHit;
				combInside = newInside;
				inter.hit[iC].isEntering = combInside;
				if (combInside) {
					nextHit.hitNormal.flip();
					inter.hit[iC].hitObject = leftObj;
				}
			}
		}
	}
	while (iL < lftInter.numHits) {
		nextHit = lftInter.hit[iL++];
		lftInside = !lftInside;
		inter.hit[++iC] = nextHit;
		inter.hit[iC].isEntering = lftInside;
	}
	inter.numHits = iC+1;
	return (inter.numHits > 0);
}
//<<<<<<<<<<<<<<<<<< hit() for Square >>>>>>>>>>>>>>>>>>>>>>
bool Plane::hit(Ray &r, Intersection &inter) {
	Ray genRay;	// generic ray
	inter.numHits = 0;
	genRay = r.xfrmRay(invTransf);
	double denom = genRay.dir.z;	// denominator
	if (fabs(denom) < 0.0001) return false;	// ray parallel to plane: miss
	double time = -genRay.start.z/denom;	// hit time
	if (time <=0.0) return false;	// it lies behind the eye
	float hx = genRay.start.x + genRay.dir.x*(float)time;	//x at hit
	float hy = genRay.start.y + genRay.dir.y*(float)time;	// y at hit
	inter.numHits = 1;	// have a hit
	inter.hit[0].hitObject = this;
	inter.hit[0].hitTime = time;
	inter.hit[0].isEntering = true;
	inter.hit[0].surface = 0;
	inter.hit[0].hitPoint.set(hx,hy,0);
	inter.hit[0].hitNormal.set(0,0,1);
	return true;
}
//<<<<<<<<<<<<<<<<<< hit() for Square >>>>>>>>>>>>>>>>>>>>>>
bool Square::hit(Ray &r, Intersection &inter) {
	Ray genRay;	// generic ray
	inter.numHits = 0;
	genRay = r.xfrmRay(invTransf);
	double denom = genRay.dir.z;	// denominator
	if (fabs(denom) < 0.0001) return false;	// ray parallel to plane: miss
	double time = -genRay.start.z/denom;	// hit time
	if (time <=0.0) return false;	// it lies behind the eye
	float hx = genRay.start.x + genRay.dir.x*(float)time;	//x at hit
	float hy = genRay.start.y + genRay.dir.y*(float)time;	// y at hit
	if ((hx > 1.0) || (hx < -1.0)) return false;	// misses in the x-direction
	if ((hy > 1.0) || (hy < -1.0)) return false;	// misses in the y-direction
	inter.numHits = 1;	// have a hit
	inter.hit[0].hitObject = this;
	inter.hit[0].hitTime = time;
	inter.hit[0].isEntering = true;
	inter.hit[0].surface = 0;
	inter.hit[0].hitPoint.set(hx,hy,0);
	inter.hit[0].hitNormal.set(0,0,1);
	return true;
}
//<<<<<<<<<<<<<<<<<< hit() for Circle >>>>>>>>>>>>>>>>>>>>>>
bool Circle::hit(Ray &r, Intersection &inter) {
	Ray genRay;	// generic ray
	inter.numHits = 0;
	genRay = r.xfrmRay(invTransf);
	double denom = genRay.dir.z;	// denominator
	if (fabs(denom) < 0.0001) return false;	// ray parallel to plane: miss
	double time = -genRay.start.z/denom;	// hit time
	if (time <=0.0) return false;	// it lies behind the eye
	float hx = genRay.start.x + genRay.dir.x*(float)time;	//x at hit
	float hy = genRay.start.y + genRay.dir.y*(float)time;	// y at hit
	if ((hx > 1.0) || (hx < -1.0)) return false;	// misses completely in the x-direction - easy outs
	if ((hy > 1.0) || (hy < -1.0)) return false;	// misses completely in the y-direction
	if ((hy >= sqrtf(1-hx*hx)) || (hy <= -sqrtf(1.0f-hx*hx)))
		return false;
	if ((hx >= sqrtf(1-hy*hy)) || (hx <= -sqrtf(1.0f-hy*hy)))
		return false;
	inter.numHits = 1;	// have a hit
	inter.hit[0].hitObject = this;
	inter.hit[0].hitTime = time;
	inter.hit[0].isEntering = true;
	inter.hit[0].surface = 0;
	inter.hit[0].hitPoint.set(hx,hy,0);
	inter.hit[0].hitNormal.set(0,0,1);
	return true;
}
//<<<<<<<<<<<<<<<<<< hit() for Triangle >>>>>>>>>>>>>>>>>>>>>>
bool Triangle::hit(Ray &r, Intersection &inter) {
	Ray genRay;	// generic ray
	inter.numHits = 0;
	genRay = r.xfrmRay(invTransf);
	double denom = genRay.dir.z;	// denominator
	if (fabs(denom) < 0.0001) return false;	// ray parallel to plane: miss
	double time = -genRay.start.z/denom;	// hit time
	if (time <=0.0) return false;	// it lies behind the eye
	float hx = genRay.start.x + genRay.dir.x*(float)time;	//x at hit
	float hy = genRay.start.y + genRay.dir.y*(float)time;	// y at hit
	if ((hx > 1.0) || (hx < -1.0)) return false;	// misses completely in the x-direction
	if ((hy > 1.0) || (hy < -1.0)) return false;	// misses completely in the y-direction
	if (hy >= -0.5 * (hx-1)) return false;	// misses in the x-direction
	if (hy <= 0.5 * (hx-1)) return false;	// misses in the x-direction
	inter.numHits = 1;	// have a hit
	inter.hit[0].hitObject = this;
	inter.hit[0].hitTime = time;
	inter.hit[0].isEntering = true;
	inter.hit[0].surface = 0;
	inter.hit[0].hitPoint.set(hx,hy,0);
	inter.hit[0].hitNormal.set(0,0,1);
	return true;
}
//<<<<<<<<<<<<<<<<<< hit() for Sphere >>>>>>>>>>>>>>>>>>>>>>>
bool Sphere::hit (Ray &r, Intersection& inter) { // intersect the ray with the generic sphere
	//if (/*doScreenExtent &&*/ r.recurseLevel == 0)
	//	if (r.col < scrnExtnt.left || r.col > scrnExtnt.right || r.row < scrnExtnt.bott || r.row > scrnExtnt.top)
	//		return false; //misses screen extent	
	Ray genRay(r.xfrmRay(invTransf));
	double A, B, C, discrim, disc_root, t1, t2;
	A = genRay.dir.dot(genRay.dir); 
	//assert(A > 0); // fails if ray is sick 
	B = genRay.start.dot3D(genRay.dir);
	C = genRay.start.dot3D(genRay.start) - 1.0;
	discrim = B * B - A * C;
	if( discrim < 0.0) return false;   // ray missed: numHits = 0 
	disc_root = sqrt(discrim);		
	inter.numHits = 0;	// no hits yet	
	t1 = (-B - disc_root)/A;         // earlier hit 
	if (t1 > 0.00001) { // a pos hit time
		inter.hit[0].hitTime = t1;
		Point3 P(genRay.rayPos(t1));// position of hit spot
		inter.hit[0].hitPoint.set(P);
		inter.hit[0].hitObject = this;
		inter.hit[0].isEntering = 1;
		inter.hit[0].surface = 0;
		inter.hit[0].hitNormal.set(P.x, P.y, P.z); // vector is same as hit spot
		inter.numHits++; // got one hit so far
	}
	t2 = (-B + disc_root)/A; 		// second hit
	if (t2 > 0.00001) { // got a hit (first or second...)
		inter.hit[inter.numHits].hitTime = t2;
		Point3 P(genRay.rayPos(t2));  // position of hit spot
		inter.hit[inter.numHits].hitPoint.set(P);
		inter.hit[inter.numHits].isEntering = 0;
		inter.hit[inter.numHits].hitObject = this;
		inter.hit[inter.numHits].surface = 0;
		inter.hit[inter.numHits].hitNormal.set(P.x, P.y, P.z); //same as hit spot
		inter.numHits++;			// got another
	}
	return inter.numHits > 0;
}// end of hit() for Sphere

//<<<<<<<<<<<<<<<<<< hit() for TaperedCylinder >>>>>>>>>>>>>>
bool TaperedCylinder::hit(Ray &r, Intersection &inter) { 
	Ray genRay; //generic ray
	genRay = r.xfrmRay(invTransf);
	double A, B, C, discrim, disc_root, t1,t2,tb,tc;
	double sm = smallRadius - 1;
	double fDir = sm * genRay.dir.z; //handy short names
	double fStart = sm * genRay.start.z + 1;
	//A = sqrtf(genRay.dir.x)+ sqrtf(genRay.dir.y)- sqrtf((float)fDir);
	A = genRay.dir.x*genRay.dir.x + genRay.dir.y*genRay.dir.y - (float)(fDir*fDir);
	B = genRay.dir.x*genRay.start.x + genRay.dir.y*genRay.start.y -(float)(fDir*fStart);
	//C = sqrtf(genRay.start.x)+ sqrtf(genRay.start.y)- sqrtf((float)fStart);
	C = genRay.start.x*genRay.start.x + genRay.start.y*genRay.start.y - (float)(fStart*fStart);
	discrim = B*B - A*C;
	int num = 0;	// no hits yet
	if (discrim > 0) {	// can take square root
		disc_root = (double)sqrt(double(discrim));
		t1 = (-B - disc_root)/A;
		float zHit = genRay.start.z + genRay.dir.z*(float)t1;	// z-component of ray
		if (t1 > 0.00001 && zHit <= 1.0 && zHit >= 0) {
			inter.hit[num].hitTime = t1;
			inter.hit[num++].surface = 0;	// hit is with wall
		}
		t2 = (-B + disc_root)/A;
		zHit = genRay.start.z + genRay.dir.z*(float)t2;	// z-component of ray
		if (t2 > 0.00001 && zHit <= 1.0 && zHit >= 0) {
			inter.hit[num].hitTime = t2;
			inter.hit[num++].surface = 0;	// hit is with wall
		}
	}	// else if (discrim > 0)
	// DOING OPEN ENDED CYLINDERS - USE CIRCLES TO CLOSE IF NEEDED
	if (num == 0) return false;	// missed everything
	inter.numHits = num;
	if (num == 1) { // eye inside cylinder, only have the exiting hit
		inter.hit[0].isEntering = false;
		inter.hit[0].hitObject = this;
	} else {	// have two hits
		if (inter.hit[0].hitTime > inter.hit[1].hitTime) {	// must reverse order
			double tmpT = inter.hit[0].hitTime;
			inter.hit[0].hitTime = inter.hit[1].hitTime;
			inter.hit[1].hitTime = tmpT;
			int tmpS = inter.hit[0].surface;
			inter.hit[0].surface = inter.hit[1].surface;
			inter.hit[1].surface = tmpS;
		}
		inter.hit[0].isEntering = true;
		inter.hit[1].isEntering = false;
		inter.hit[0].hitObject = inter.hit[1].hitObject = this;
	}
	// now set the hit point and normal for the hit or hits
	for (int i = 0; i<num; i++) {
		Point3 P0(genRay.rayPos(inter.hit[i].hitTime));	// position of the first hit
		inter.hit[i].hitPoint.set(P0);
		int surf = inter.hit[i].surface;
		if (surf == 0)	// wall
			inter.hit[i].hitNormal.set(P0.x, P0.y, -(float)sm*(1+(float)sm*P0.z));
		else if (surf == 1)	// base
			inter.hit[i].hitNormal.set(0, 0, -1);
		else inter.hit[i].hitNormal.set(0, 0, 1);	// cap
	}
	return true;
}
//<<<<<<<<<<<<<<<<<<<<<<<<<<<< readMesh >>>>>>>>>>>>>>>>>>>>>>>>
void Mesh::readMesh(string fname) {
	fstream inStream;
	inStream.open(fname.c_str(), ios ::in); //open needs a c-like string
	if(inStream.fail() || inStream.eof()) {
		cout << "can't open file or eof: " << fname << endl; 
		makeEmpty();
		return;
	}
	inStream >> numVerts >> numNorms >> numFaces;
	// make arrays for vertices, normals, and faces
	pt = new Point3[numVerts];        assert(pt != NULL);
	norm = new Vector3[numNorms];     assert(norm != NULL);
	face = new Face[numFaces];        assert(face != NULL);
	for(int i = 0; i < numVerts; i++) 	// read in the vertices
		inStream >> pt[i].x >> pt[i].y >> pt[i].z;
	for(int ii = 0; ii < numNorms; ii++) 	// read in the normals
		inStream >> norm[ii].x >> norm[ii].y >> norm[ii].z;
	for(int f = 0; f < numFaces; f++) { // read in face data
		inStream >> face[f].nVerts;
		int n = face[f].nVerts;
		face[f].vert = new VertexID[n]; assert(face[f].vert != NULL);
		for(int k = 0; k < n; k++) 		// read vertex indices for this face
			inStream >> face[f].vert[k].vertIndex;
		for(int kk = 0; kk < n; kk++) 		// read normal indices for this face
			inStream >> face[f].vert[kk].normIndex;
	}
	inStream.close();
} // end of readMesh

//<<<<<<<<<<<<<<<<<<<<<< drawMesh >>>>>>>>>>>>>>>>>>>>
void Mesh :: drawMesh() { // draw each face of this mesh using OpenGL: draw each polygon.
	if(isEmpty()) return; // mesh is empty
	for(int f = 0; f < numFaces; f++) {
		int n = face[f].nVerts;
		glBegin(GL_POLYGON);
		for(int v = 0; v < n; v++) {
			int in = face[f].vert[v].normIndex;
			assert(in >= 0 && in < numNorms);
			glNormal3f(norm[in].x, norm[in].y, norm[in].z);
			int iv = face[f].vert[v].vertIndex; assert(iv >= 0 && iv < numVerts);
			glVertex3f(pt[iv].x, pt[iv].y, pt[iv].z);
		}
		glEnd();	
	}
	glFlush();
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<< write mesh>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
void Mesh:: writeMesh(char * fname) { // write this mesh object into a new Chapter 6 format file.
	if(numVerts == 0 || numNorms   == 0 || numFaces == 0) return; //empty
	fstream outStream(fname, ios ::out); // open the output stream
	if(outStream.fail()) {cout << "can't make new file: " << fname << endl; return;}
	outStream << numVerts << " " << numNorms << " " << numFaces << "\n";
	// write the vertex and vertex normal list
	for(int i = 0; i < numVerts; i++)
		outStream << pt[i].x   << " " << pt[i].y   << " " << pt[i].z << "\n";
	for(int ii = 0; ii < numNorms; ii++)
		outStream  << norm[ii].x << " " << norm[ii].y << " " << norm[ii].z << "\n";
	// write the face data
	for(int f = 0; f < numFaces; f++) {
		int n = face[f].nVerts;
		outStream << n << "\n";
		for(int v = 0; v < n; v++)// write vertex indices for this face
			outStream << face[f].vert[v].vertIndex << " ";	outStream << "\n";
		for(int k = 0; k < n; k++)	// write normal indices for this face 
			outStream << face[f].vert[k].normIndex << " "; outStream << "\n";
	}
	outStream.close();
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<drawFrontFaces>>>>>>>>>>>>>>>>>>>>>>>>>>
void Mesh:: drawFrontFaces() { // use OpenGL to draw this mesh            
	for (int f = 0; f < numFaces; f++) { // draw each face
		if (false/*isBackFace(f,)*/) continue;
		glBegin(GL_POLYGON);
			for (int v = 0; v < face[f].nVerts; v++) { // for each one..
				int in = face[f].vert[v].normIndex ; // index of this normal
				int iv =  face[f].vert[v].vertIndex ; // index of this vertex
				glNormal3f(norm[in].x, norm[in].y, norm[in].z);
				glVertex3f(pt[iv].x, pt[iv].y, pt[iv].z);
			}
		glEnd();
	}
}

//<<<<<<<<<<<<<<< Scene :: read >>>>>>>>>>>>>>>>
bool Scene::read(string fname) { // return true if ok; else false
	file_in = new ifstream(fname.c_str());
	if(!(*file_in)) { 
		cout << "I can't find or open file: " << fname << endl; 
		return false;
	}
	f_in = new strstream();
	line = nextline = 1;
	def_stack = new DefUnitStack();
	char ch;
	freeScene(); //delete any previous scene
	
	// initialize all for reading:
	obj = tail = NULL;
	light = NULL;
	affStk.tos = new AffineNode;
	affStk.tos->next = NULL;
	
	while (file_in->get(ch)) {
		*f_in << ch;
	} // read whole file
	while(1) { //read file, collecting objects, until EOF or an error
		GeomObj * shp = getObject(); // get the next shape
		if(!shp) break; // no object: either error or EOF
		shp->next = NULL; // to be safe
		if(obj == NULL){ obj = tail = shp;} // empty list so far
		else{tail->next = shp; tail = shp;} // add new object to queue
	}	
	file_in->close();
	cleanUp(); // delete temp lists, etc.
	return true;
} // end of read()

//<<<<<<<<<<<<<< Scene :: getObject >>>>>>>>>>>>>>>	
GeomObj* Scene :: getObject() {
	//reads tokens from stream f_in (a data member of Scene),
	// building lights, getting materials, doing transformations,
	// until it finds a new object
	// returns NULL if any error occurs, or end of file
	string s;			
	GeomObj * newShape;
	TokenType typ;
	while ((typ = (whichtoken( s = nexttoken() ))) != T_NULL) {
		if(typ == UNION || typ == INTERSECTION || typ == DIFFERENCEa) {
			switch(typ) {
			case UNION:			newShape = new UnionBool();			break;
			case INTERSECTION:	newShape = new IntersectionBool();	break;
			case DIFFERENCEa:	newShape = new DifferenceBool();	break;
			} // end of little switch
			GeomObj* p = newShape;
			p = getObject(); // get left child
			if(!p) return NULL; // Error! should always get an object
			((Boolean*)newShape)->left  = p; // hook it up
			p = getObject();// get right child
			if(!p) return NULL;
			((Boolean*)newShape)->right = p; // hook it up
			return newShape;
		}// end of if(typ == UNION etc....
		switch(typ) {
		case LIGHT: {
			Point3 p;
			Color3 c;
			p.x = getFloat(); p.y = getFloat();	p.z = getFloat();
			c.red = getFloat(); c.green = getFloat();	c.blue = getFloat();
			Light *l = new Light;
			l->setPosition(p);
			l->setColor(c);
			l->next = light; //put it on the list
			light = l; break;}
		case ROTATE: {
			float angle;
			Vector3 u;
			angle = getFloat(); u.x = getFloat(); 
			u.y = getFloat(); u.z = getFloat();
			affStk.rotate(angle,u);break;}
		case TRANSLATE: {
			Vector3 d;
			d.x = getFloat(); d.y = getFloat(); d.z = getFloat();
			affStk.translate(d);break;}
		case SCALE: {
			float sx, sy, sz;
			sx = getFloat(); sy = getFloat(); sz = getFloat();
			affStk.scale(sx, sy, sz);break;}
		case PUSH: affStk.dup(); break; 
		case POP:  affStk.popAndDrop(); break;
		case IDENTITYAFFINE: affStk.setIdentity();break;		
		case AMBIENT: {
			float dr, dg, db;
			dr = getFloat(); dg = getFloat(); db = getFloat();
			currMtrl.ambient.set(dr,dg,db); break;}
		case DIFFUSE: {
			float dr,dg,db;
			dr = getFloat(); dg = getFloat(); db = getFloat();
			currMtrl.diffuse.set(dr,dg,db); break;}
		case SPECULAR:{ 
			float dr,dg,db;
			dr = getFloat(); dg = getFloat(); db = getFloat();
			currMtrl.specular.set(dr,dg,db); break;}
		case EMISSIVE: {
			float dr,dg,db;
			dr = getFloat(); dg = getFloat(); db = getFloat();
			currMtrl.emissive.set(dr,dg,db); break;}
		case PARAMETERS: { // get a list of numParams parameters
			currMtrl.numParams = (int)getFloat();
			for(int i = 0; i < currMtrl.numParams; i++)
				currMtrl.params[i] = getFloat();
			break;}			
		case SPECULARFRACTION: currMtrl.specularFraction = getFloat(); break;
		case SURFACEROUGHNESS: currMtrl.surfaceRoughness = getFloat(); break;
		case TEXTURE: { // get type, 0 for none
				currMtrl.textureType = (int)getFloat();	
				break;}
		case DEFAULTMATERIALS: 	currMtrl.setDefault();break;
		case SPEEDOFLIGHT: currMtrl.speedOfLight = getFloat(); break;
		case SPECULAREXPONENT: currMtrl.specularExponent = getFloat(); break;
		case TRANSPARENCY:currMtrl.transparency = getFloat(); break;
		case REFLECTIVITY: currMtrl.reflectivity = getFloat(); break;
		case GLOBALAMBIENT:
			ambient.red = getFloat();
			ambient.green = getFloat();
			ambient.blue = getFloat();
			break;
		case BACKGROUND: 
			background.red = getFloat();
			background.green = getFloat(); 
			background.blue = getFloat();
			break;
		case MINREFLECTIVITY: minReflectivity = getFloat(); break;
		case MINTRANSPARENCY: minTransparency = getFloat(); break;
		case MAXRECURSIONDEPTH: maxRecursionDepth = (int)getFloat(); break;
		case MAKEPIXMAP: {	// get BMP file name for a pixmap			
		  int which = (int)getFloat();// index of this pixmap in pixmap array
		  if(which < 0 || which > 7){cout << "\nbad index of RGBpixmap!\n";}
		  string fname = nexttoken(); // get file name for mesh	  
		  cout << "I got fname = " << fname << endl;
		  if(!pixmap[which].readBMPFile(fname))
		  {// read BMP file into this pixmap
				cout << " \ncan't read that RGBpixmap file!\n";
				return NULL;
		  }
		  break;}// end of case: MAKEPIXMAP
		case T_NULL: break; // The null token represents end-of-file 
		case DEF: {
			string name, temp, lb, rb;
			int l = line;
			string inp;
			name = nexttoken();
			if ( whichtoken(name) != IDENT ) {
				cout << "Error:  Identifier expected." << endl;
				return NULL;
			}
			if ( def_stack->search(name) ) {
				cout << line << ": " << name;
				cout << ": attempt to redefine. " << endl;
				return NULL;
			}
			lb = nexttoken();
			if ( whichtoken(lb) != LFTCURLY ) {
				cout << "Error: { expected." << endl;
				return NULL;
			}
			while ( whichtoken( temp = nexttoken()) != RGHTCURLY ) {
				cout << temp << endl;
				inp = inp + temp + " ";
				if (!f_in) {
					cout << "Error: end of file detected." << endl;
					return NULL;
				}
			}
			// Push the contents of the string on the stack.
			def_stack->push(name, inp);
			break;} // end of case: DEF
		case USE: {
			string name;
			name = nexttoken();
			if ( whichtoken(name) != IDENT ) {
				cout << line << ": " << name;
				cout << ": identifier expected.";
				return NULL;
			}
			if (! def_stack->search(name) ) {
				cout << line << ": " << name;
				cout << ": not defined.";
				return NULL;
			}
			cout << def_stack->contents(name) << endl;
			strstream *temp_fin = new strstream;
			*temp_fin << def_stack->contents(name) << " ";
			*temp_fin << f_in->rdbuf();
			delete (f_in);
			f_in = temp_fin;
			break; } // end of case: USE
		default:  { // inner switch for Shapes
			switch(typ) 
			{
				//float param;
			case CUBE:				newShape = new Cube;				break;
			case SPHERE:			newShape = new Sphere;				break;
			case CIRCLE:			newShape = new Circle;				break;
			case TETRAHEDRON:		newShape = new Mesh("tetra.3vn");	break;
			//case TORUS:			newShape = new Torus;				break;
			case PLANE:				newShape = new Plane;				break;
			case SQUARE:			newShape = new Square;				break;
			case TRIANGLE:			newShape = new Triangle;			break;
			case TAPEREDCYLINDER:	newShape = new TaperedCylinder; 
				((TaperedCylinder*)newShape)->smallRadius = getFloat();	break;
			case CONE:				newShape = new TaperedCylinder; 
				((TaperedCylinder*)newShape)->smallRadius = 0;			break;
			case CYLINDER:			newShape = new TaperedCylinder; 
				((TaperedCylinder*)newShape)->smallRadius = 1;			break;
			case OCTAHEDRON:		newShape = new Mesh("octa.3vn");	break;
			case DODECAHEDRON:		newShape = new Mesh("dodeca.3vn");	break;
			case ICOSAHEDRON:		newShape = new Mesh("icosa.3vn");	break;
			case BUCKYBALL:			newShape = new Mesh("bucky.3vn");	break;
			case DIAMOND:			newShape = new Mesh("diamond.3vn");	break;
			//case TEAPOT:			newShape = new Teapot;	break;
			case MESH: {// get a filename (with extension) for this mesh		
				string fname = nexttoken(); // get file name for mesh
				newShape = new Mesh(fname); break;
				}// end of case: MESH
			default: {
				cerr << "Line " << nextline << ": unknown keyword " << s << endl;
				return NULL;
					 }	
			} // end of inner switch
			// common things to do to all Shapes
			((Shape*)newShape)->mtrl.set(currMtrl);
			// load transform and its inverse
			((Shape*)newShape)->transf.set(*(affStk.tos->affn));
			((Shape*)newShape)->invTransf.set(*(affStk.tos->invAffn));
			return newShape;
		}// end of default: block			
		} // end of outer switch
	} // end of while
	return NULL;
} // end of getObject

//<<<<<<<<<<<<<<< Scene :: shade >>>>>>>>>>>>>>>
Color3 Scene::shade(Ray &ray) {	// return the color of this ray
	float epsilon = .000001f;
	Intersection best;	// data for the best hit so far
	getFirstHit(ray, best);	// fill the 'best' record
	if (best.numHits == 0)	// did the ray miss every object?
		return background;
	HitInfo h = best.hit[0]; //data about the first hit

	// get the hit point
	Point3 hitPoint(ray.start.x + (ray.dir.x * (float)h.hitTime), 
					ray.start.y + (ray.dir.y * (float)h.hitTime),
					ray.start.z + (ray.dir.z * (float)h.hitTime));
	// direction to viewer
	Vector3 v(-ray.dir.x, -ray.dir.y, -ray.dir.z);
	v.normalize();
	
	// pointer to hit object
	Shape* myObj = (Shape*)h.hitObject;

	// add emissive, ambient components
	Color3 color(myObj->mtrl.emissive);
	color.add(ambient, myObj->mtrl.ambient);
	Color3 reflectColor, refractColor, diffuseColor, specColor; // component colors for later

	// calculate the normal vector
	Vector3 normal;
	xfrmNormal(normal,myObj->invTransf,h.hitNormal); //transform the generic normal to the world normal
	normal.normalize();

	// calculate direction of ray from eye
	Vector3 dir = ray.dir;
	dir.normalize();

	// init shadow feelers
	Ray feeler;
	Point3 feelerOrigin(hitPoint.x-(epsilon*dir.x), hitPoint.y-(epsilon*dir.y), hitPoint.z-(epsilon*dir.z));
	feeler.setStart(feelerOrigin);
	feeler.recurseLevel = 1;

	// calculate lighting via phong lighting model
	for (Light *l = light; l; l=l->next) {
		// shadow calculation
		feeler.setDir(l->pos.x-hitPoint.x, l->pos.y-hitPoint.y, l->pos.z-hitPoint.z);
		feeler.dir.normalize();
		if(isInShadow(feeler)) continue;

		// lambert color calculation
		Vector3 s(l->pos.x-hitPoint.x, l->pos.y-hitPoint.y, l->pos.z-hitPoint.z); // vector from hit to source
		s.normalize();
		float mDotS = s.dot(normal); // the Lambert term

		if (mDotS > 0.0) { //hit point is turned towards light
			diffuseColor.red = mDotS * myObj->mtrl.diffuse.red * l->color.red;
			diffuseColor.green = mDotS * myObj->mtrl.diffuse.green * l->color.green;
			diffuseColor.blue = mDotS * myObj->mtrl.diffuse.blue * l->color.blue;
			color.add(diffuseColor);
			Vector3 h(v.x+s.x, v.y+s.y, v.z+s.z);
			h.normalize();
			float mDotH = h.dot(normal); // part of phong term
			if (mDotH > 0.0) {
				float phong = pow(mDotH, myObj->mtrl.specularExponent);
				specColor.red = phong * myObj->mtrl.specular.red * l->color.red;
				specColor.green = phong * myObj->mtrl.specular.green * l->color.green;
				specColor.blue = phong * myObj->mtrl.specular.blue * l->color.blue;
				color.add(specColor);
			}
		}
	}
	// exit from recursion if necessary
	if (ray.recurseLevel >= maxRecursionDepth)
		return color;

	// calculate reflection
	if (myObj->mtrl.reflectivity > 0) {
		Ray refl;
		refl.setStart(hitPoint);
		dir.set(ray.dir);
		dir.normalize();
		float temp = 2*dir.dot(normal);
		Vector3 tV(normal);
		tV.x *= temp;
		tV.y *= temp;
		tV.z *= temp;

		tV.flip();
		refl.dir.x += tV.x;
		refl.dir.y += tV.y;
		refl.dir.z += tV.z;
		refl.dir.normalize();

		refl.recurseLevel = ray.recurseLevel + 1;
		reflectColor = shade(refl);

		reflectColor.red *= myObj->mtrl.reflectivity;
		reflectColor.green *= myObj->mtrl.reflectivity;
		reflectColor.blue *= myObj->mtrl.reflectivity;
		color.add(reflectColor);
	}

	// calculate transmission
	if (myObj->mtrl.transparency > 0) {
		float c1 = 1.0, c2 = myObj->mtrl.speedOfLight; //speed of light in air and in material, respectively
		Point3 p(ray.rayPos(best.hit[0].hitTime+epsilon));
		Ray refr;
		refr.setStart(p);
		refr.recurseLevel = ray.recurseLevel + 1;
		refr.dir.set(ray.dir);
		refr.dir.normalize();
		float temp1 = dir.dot(normal);
		float r;
		if (temp1 >= 0) {
			normal.flip();
			r = c1/c2;
		} else r = c2/c1;

		float f = refr.dir.dot(normal);
		float t = 1 - r*r*(1-f*f);
		if (t < 0) { // total internal reflection
			color.set(0,0,0);
			return color;
		}
		float cosTheta = sqrt(t);
		Vector3 temp(normal);
		temp.x *= (r*f-cosTheta);
		temp.y *= (r*f-cosTheta);
		temp.z *= (r*f-cosTheta);
		refr.dir.x *= r;
		refr.dir.y *= r;
		refr.dir.z *= r;
		refr.dir.x += temp.x;
		refr.dir.y += temp.y;
		refr.dir.z += temp.z;
		refr.dir.normalize();
		refractColor = shade(refr);
		refractColor.red *= myObj->mtrl.transparency;
		refractColor.green *= myObj->mtrl.transparency;
		refractColor.blue *= myObj->mtrl.transparency;
		color.add(refractColor);
	}
	return color;
}
//<<<<<<<<<<<<<< scene :: getFirstHit >>>>>>>>>>>>>>>>>
void Scene::getFirstHit(Ray& ray, Intersection& best) {
	Intersection inter;		// make the intersection record
	best.numHits=0;			// no hits yet
	for (GeomObj* pObj=obj; pObj != NULL; pObj = pObj->next) {	// test each object in the scene
		if (!pObj->hit(ray, inter))
			continue;
		if (best.numHits == 0 || inter.hit[0].hitTime < best.hit[0].hitTime)
			best.set(inter);	// copy inter into best
	}
}
//<<<<<<<<<<<<<< scene :: IsInShadow() >>>>>>>>>>>>>>>
bool Scene::isInShadow(Ray &f) {
	Intersection tmp;
	for (GeomObj* p = obj; p; p = p->next)
		if (p->hit(f, tmp)) return true;
	return false;
}