// Shapes class and Supporting classes
// Shapes.h 
// GeomObj, Boolean, Shape, Cube, Sphere, etc.
// uses pixmap class and basic classes above

#pragma once
#include <string>
#define PI 3.14159265
#define TWOPI (3.14159265 * 2)
#define degPerRad (180/3.14159265)
#define Near(a,v) fabs((a)-(v))<0.0001

//@@@@@@@@@@@@@@@@@ Material class @@@@@@@@@@@@@@
class Material {
public:
	Color3 ambient, diffuse, specular, emissive;
	int numParams; // for textures
	float params[10]; // for textures
	int textureType;  // 0 for none, neg for solids, pos for images
	float specularExponent, reflectivity, transparency, speedOfLight;
	float specularFraction, surfaceRoughness;	
	void setDefault() {
		textureType = 0; // for none
		numParams = 0;
		reflectivity = transparency = 0.0;
		speedOfLight = specularExponent = 1.0;
		specularFraction = 0.0;
		surfaceRoughness = 1.0;
		ambient.set(0.1f,0.1f,0.1f);
		diffuse.set(0.8f,0.8f,0.8f);
		specular.set(0,0,0);
		emissive.set(0,0,0);
	}
	void set(Material& m) {
		textureType = m.textureType;
		numParams = m.numParams;
		for(int i = 0; i < numParams; i++) params[i] = m.params[i];
		transparency = m.transparency;
		speedOfLight = m.speedOfLight;
		reflectivity = m.reflectivity;
		specularExponent = m.specularExponent;
		specularFraction = m.specularFraction;
		surfaceRoughness = m.surfaceRoughness;
		ambient.set(m.ambient);
		diffuse.set(m.diffuse);
		specular.set(m.specular);
		emissive.set(m.emissive);
	}
}; // end of Material

//@@@@@@@@@@@@@@@@@@@@@ GeomObj class @@@@@@@@@@@@@@@@
class GeomObj {
public: 
		IntRect scrnExtnt;
		Cuboid genBoxExtent,worldBoxExtent;
		SphereInfo genSphereExtent,worldSphereExtent;
		GeomObj * next;
		GeomObj(): next(NULL) {}
		virtual bool hit(Ray &r, Intersection &inter) { return false; }
		virtual void loadStuff() {}
		virtual void drawOpenGL() {}
		virtual void tellMaterialsGL() {}
		virtual void makeExtentPoints(PointCluster& clust) { cout << "in makeExtentPoints for GeomObj"; }
		virtual Point2 texturePoint(Point3 p) { Point2 pt(0,0); return pt; }
		virtual Color3 texture(HitInfo& h) { Color3 c(0,0,0); return c; }
};

//@@@@@@@@@@@@@@@@@@@ Boolean @@@@@@@@@@@@@@@@@@
class Boolean: public GeomObj {
public:
	GeomObj *left, *right;
	Boolean():left(NULL),right(NULL){}
	virtual bool hit(Ray &r, Intersection &inter){return false;}
	virtual void makeExtentPoints(PointCluster& clust) { cout << "in makeExtentPoints for Boolean"; }
	virtual void drawOpenGL() { // just draw its children
		if(left)left->drawOpenGL();
		if(right)right->drawOpenGL();
	}
};

//@@@@@@@@@@@@@@@@@@@@ UnionBool @@@@@@@@@@@@@@@@
class UnionBool : public Boolean {
public:
	UnionBool() { Boolean(); } //constructor
	virtual void makeExtentPoints(PointCluster& clust) { // cluster is union of lft and rt clusters
		PointCluster lftClust, rtClust;
		left->makeExtentPoints(lftClust);
		right->makeExtentPoints(rtClust);
		// now combine them
		int i,j;
		clust.num = lftClust.num + rtClust.num;
		clust.pt = new Point3[clust.num];
		assert(clust.pt);
		for(i = 0; i < lftClust.num; i++)
			clust.pt[i] = lftClust.pt[i];
		for(j = 0; j < rtClust.num; j++)
			clust.pt[j + i] = rtClust.pt[j];
	}
	virtual bool hit(Ray &r, Intersection &inter);
};

//@@@@@@@@@@@@@@@@@@@@ IntersectionBool @@@@@@@@@@@@@@@@
class IntersectionBool : public Boolean {
public:
	IntersectionBool() { Boolean(); }
	virtual void makeExtentPoints(PointCluster& clust) { // cluster is just left cluster	
		left->makeExtentPoints(clust);
	}
	virtual bool hit(Ray &r, Intersection &inter);
};

//@@@@@@@@@@@@@@@@@@@@ DifferenceBool @@@@@@@@@@@@@@@@
class DifferenceBool : public Boolean {
public:
	DifferenceBool() { Boolean(); } 
	virtual void makeExtentPoints(PointCluster& clust) { // cluster is just left cluster
		left->makeExtentPoints(clust);
	}
	virtual bool hit(Ray &r, Intersection &inter);
};

//@@@@@@@@@@@@@@@@@ Shape @@@@@@@@@@@@@@@@@@@@@
class Shape: public GeomObj {
public:
	Material mtrl;
	Affine4 transf,invTransf;
	virtual bool hit(Ray &r, Intersection &inter) { return false; }
	//virtual Color3 texture(HitInfo& h, int whichTexture);
	Shape(){mtrl.textureType = 0; mtrl.numParams = 0;}
	void setMaterial(Material& mt){mtrl.set(mt);}
	void tellMaterialsGL() {
		float amb[4],diff[4],spec[4], emiss[4];
		float zero[] = {0,0,0,1};
		mtrl.ambient.build4tuple(amb); // fill the array
		mtrl.diffuse.build4tuple(diff);
		mtrl.specular.build4tuple(spec);
		mtrl.emissive.build4tuple(emiss);
		glMaterialfv(GL_FRONT/*_AND_BACK*/,GL_AMBIENT,amb);
		glMaterialfv(GL_FRONT/*_AND_BACK*/,GL_DIFFUSE,diff);
		if(specularOn)
			glMaterialfv(GL_FRONT/*_AND_BACK*/,GL_SPECULAR,spec);
		else 
			glMaterialfv(GL_FRONT,GL_SPECULAR,zero);
		glMaterialfv(GL_FRONT/*_AND_BACK*/,GL_EMISSION,emiss);
		glMaterialf (GL_FRONT/*_AND_BACK*/,GL_SHININESS,mtrl.specularExponent);
	}
	virtual void drawOpenGL() {}
	virtual Point2 texturePoint(Point3 p) { Point2 pt(0,0); return pt; }
	virtual void makeExtentPoints(PointCluster& clust) {}		
}; //end: Shape class

 
//@$@$@$@$@$@$@$@$@ Cube class $@$@$@$@$@$@$@$@$@$@
class Cube : public Shape {
public:	
	Cube() {}
	void drawOpenGL() {
		tellMaterialsGL(); glPushMatrix();
		glMultMatrixf(transf.m); //load affine
		glEnable(GL_NORMALIZE);
		if(doEdges) glutWireCube(2.0);
		if(doFaces) glutSolidCube(2.0);	// a cuber with vertices -1 to +1
		glPopMatrix();
	}
	bool hit(Ray &r, Intersection &inter); // implementation left to the reader
	Point2 texturePoint(Point3 p); // ditto
	void makeExtentPoints(PointCluster& clust); // ditto
};

//@$@$@$@$@$@$@$@$ Sphere class @$@$@$@$@$@$@$@$@$@$@$@
class Sphere : public Shape {
public:
	void drawOpenGL() { 
		tellMaterialsGL(); 
		glPushMatrix();	
			glMultMatrixf(transf.m); 
			glutSolidSphere(1.0,20,20);	
		glPopMatrix();
	}
	Sphere() {}
	bool hit(Ray& r, Intersection& inter);
	Point2 texturePoint(Point3 p);
	void makeExtentPoints(PointCluster& clust);
};
//@$@$@$@$@$@$@$@$@$@ TaperedCylinder class @$@$@$@$@$@$@$@$@$
class TaperedCylinder : public Shape {
public:
	float smallRadius;
	TaperedCylinder() {}
	bool hit(Ray &r, Intersection &inter);		
	void drawOpenGL();
	Point2 texturePoint(Point3 p);
	void makeExtentPoints(PointCluster& clust);
};

//@$@$@$@$@$@$@$@$@$@ Square class @$@$@$@$@$@$@$@$@$
class Square : public Shape {
public:
	Square() {}
	bool hit(Ray &r, Intersection &inter);
	void drawOpenGL();
	Point2 texturePoint(Point3 p);
	void makeExtentPoints(PointCluster& clust);
};

//@$@$@$@$@$@$@$@$@$@ Plane class @$@$@$@$@$@$@$@$@$
class Plane : public Shape {
public:
	Plane() {}
	bool hit(Ray &r, Intersection &inter);
	void drawOpenGL();
	Point2 texturePoint(Point3 p);
	void makeExtentPoints(PointCluster& clust);
};

//################## class VertexID ################
//used to define a Mesh
class VertexID { public: int vertIndex, normIndex; };
//################# class FACE ##############
//used to define a Mesh
class Face {
public:
	int nVerts;
	VertexID * vert; // array of vertex and normal indices
	Face(){ nVerts = 0; vert = NULL;}
	~Face(){delete[] vert; nVerts = 0;}
};
//@$@$@$@$@$@$@$@$@$@ Mesh class @$@$@$@$@$@$@$@$@$
class Mesh : public Shape {
private:
	int numVerts, numNorms, numFaces;
	Point3 *pt; // array of points
	Vector3 *norm; // array of normals
	Face *face; // array of faces
	int lastVertUsed;
	int lastNormUsed;
	int lastFaceUsed;
public:
	void readMesh(string fname);
	void writeMesh(char* fname);
	void printMesh();
	void drawMesh();
	void drawEdges();
	void freeMesh() { // free up memory used by this mesh.
		delete [] pt; // release whole vertex list
		delete [] norm;
		for(int f = 0; f < numFaces; f++)
			delete[] face[f].vert; // delete the vert[] array of this face
		delete [] face;
	}
	int isEmpty() { return (numVerts == 0) || (numFaces == 0) || (numNorms == 0); }
	void makeEmpty() { numVerts = numFaces = numNorms = 0; }
	Mesh() {
		numVerts = numFaces = numNorms = 0; 
		pt = NULL; norm  =  NULL; face = NULL;
		lastVertUsed = lastNormUsed = lastFaceUsed = -1;
	}
	virtual void drawOpenGL() { 
		tellMaterialsGL();
		glPushMatrix();
		glMultMatrixf(transf.m); 
		if (doFaces) drawMesh();	
		if (doEdges) drawEdges();
		glPopMatrix();
	} 
	Mesh(string fname){ // read this file to build mesh
		numVerts = numFaces = numNorms = 0; 
		pt = NULL; norm  =  NULL; face = NULL;
		lastVertUsed = lastNormUsed = lastFaceUsed = -1;
		readMesh(fname);
	}
	Vector3 newell4(int indx[]) {
		/* return the normalized normal to face with vertices
		pt[indx[0]],...,pt[indx[3]]. i.e. indx[] contains the four indices 
		into the vertex list to be used in the Newell calculation */
		Vector3 m;
		for(int i = 0; i < 4 ; i++) {
			int next = (i== 3) ? 0 : i + 1; // which index is next?
			int f = indx[i], n = indx[next]; // names for the indices in the pair
			m.x += (pt[f].y - pt[n].y) * (pt[f].z + pt[n].z);
			m.y += (pt[f].z - pt[n].z) * (pt[f].x + pt[n].x);
			m.z += (pt[f].x - pt[n].x) * (pt[f].y + pt[n].y);
		}
		m.normalize();
		return m;
	}
public:
	string meshFileName; // holds file name for this Mesh
	void print() { cout << "\nMesh ";mtrl.print(); }
	bool hit(Ray &r, Intersection &inter);
	Point2 texturePoint(Point3 p);
	void makeExtentPoints(PointCluster& clust);
};

//@$@$@$@$@$@$@$@$@$@ Torus class @$@$@$@$@$@$@$@$@$
class Torus : public Shape {
public:
	bool hit(Ray &r, Intersection &inter){return false;}
	void drawOpenGL() { 
		tellMaterialsGL();
		glPushMatrix();
			glMultMatrixf(transf.m); 
			if(doFaces) glutSolidTorus(0.2,1.0,10,12);	
			if(doEdges) glutWireTorus(0.2,1.0,10,12);	
		glPopMatrix();
	}
};

//@$@$@$@$@$@$@$@$@$@ Teapot class @$@$@$@$@$@$@$@$@$
class Teapot : public Shape {
public:
	bool hit(Ray &r, Intersection &inter){return false;}
	void drawOpenGL(){
		tellMaterialsGL();
		glPushMatrix();
			glMultMatrixf(transf.m); 
			glutSolidTeapot(1.0);
		glPopMatrix();
	}
};
