#pragma once
#include <iostream>
#include <assert.h>
#include <math.h>
using namespace std;
class Vector3;
class Affine4;
//@@@@@@@@@@@@@@@@@@ IntPoint class @@@@@@@@@@@@@@@@
class IntPoint{ // for 2D points with integer coordinates
public: 
	int x,y;
	void set(int dx, int dy){x = dx; y = dy;}
	void set(IntPoint& p){ x = p.x; y = p.y;} 
	IntPoint(int xx, int yy){x = xx; y = yy;}
	IntPoint(){ x = y = 0;}
};
//@@@@@@@@@@@@@@@@@@ Point2 class @@@@@@@@@@@@@@@@
class Point2{ // for 2D points with real coordinates
public: 
	float x,y;
	void set(float dx, float dy){x = dx; y = dy;}
	void set(Point2& p){ x = p.x; y = p.y;} 
	Point2(float xx, float yy){x = xx; y = yy;}
	Point2(){x = y = 0;}
};
//<<<<<<<<<<<<<<<<<<<<<< PolyLine >>>>>>>>>>>>>>>>>>>>>>>>>
class PolyLine{ // a polyline is a num plus an array of points
public:
	int num;
	Point2 pt[80]; //may need larger arrays in some circumstances
	PolyLine(){num = 0;}
};
// @@@@@@@@@@@@@@@@@@@@@@@@ IntRect class @@@@@@@@@@@@@@@@@@@@
class IntRect{ // a rectangle with integer border values
public: 
  int left, top, right, bott;
  IntRect(){left = top = right = bott = 0;}
  IntRect(int l, int t, int r, int b)
  {left = l; top = t; right = r; bott = b;}
  void set(int l, int t, int r, int b)
  {left = l; top = t; right = r; bott = b;}
  void set(IntRect& r)
  {left = r.left; top = r.top; right = r.right; bott = r.bott;}
};
//@@@@@@@@@@@@@@@@@@ Vector2 class @@@@@@@@@@@@@@@@
class Vector2{ 
public: 
	float x,y;
	void set(float dx, float dy){ x = dx; y = dy; } 
	void set(Vector2& v){ x = v.x; y = v.y;}
	void setDiff(Point2& a, Point2& b)//set to difference a - b
	{x = a.x - b.x; y = a.y - b.y;}
	void normalize()//adjust this vector to unit length
	{		double sizeSq = x * x + y * y;
		if(sizeSq < 0.0000001)
		{
			cerr << "\nnormalize() sees vector (0,0)!";
			return; // does nothing to zero vectors;
		}
		float scaleFactor = 1.0f/(float)sqrt(sizeSq);
		x *= scaleFactor; y *= scaleFactor;
	}
	Vector2(float xx, float yy){x = xx; y = yy; }
	Vector2(Vector2& v){x = v.x; y = v.y; }
	Vector2(){x = y = 0;} //default constructor
	float dot(Vector2 b) // return this dotted with b
 {return x * b.x + y * b.y;}
	void perp() // perp this vector
	{float tmp = x; x = -y; y = tmp;}
	float perpDot(Vector2& v) // return perp of this dotted with v
	{return x *v.x - y * v.y;}
};
//@@@@@@@@@@@@@@@@@@ Point3 class @@@@@@@@@@@@@@@@
class Point3 { 
public: 
	float x, y, z;
	void set(float dx, float dy, float dz) { x = dx; y = dy; z = dz; }
	void set(Point3& p) { x = p.x; y = p.y; z = p.z; }
	Point3(float xx, float yy, float zz) { x = xx; y = yy; z = zz; }
	Point3() { x = y = z = 0; }
	void build4tuple(float v[])	{ // load 4-tuple with this color: v[3] = 1 for homogeneous
		v[0] = x; v[1] = y; v[2] = z; v[3] = 1.0f;
	}
	double dot3D(Point3 b) { // return this dotted with b
		return x * b.x + y * b.y + z * b.z;
	}
	double dot3D(Vector3 b); // return this dotted with b - BELOW Vector3 class
};

//@@@@@@@@@@@@@@@@@@ Vector3 class @@@@@@@@@@@@@@@@
class Vector3 { 
public: 
	float x, y, z;
	void set(float dx, float dy, float dz) { x = dx; y = dy; z = dz; } 
	void set(Vector3& v) { x = v.x; y = v.y; z = v.z; }
	void flip() { x = -x; y = -y; z = -z; } // reverse this vector
	void setDiff(Point3& a, Point3& b) { //set to difference a - b
		x = a.x - b.x; y = a.y - b.y; z = a.z - b.z;
	}
	void normalize() { //adjust this vector to unit length		
		double sizeSq = x * x + y * y + z * z;
		if(sizeSq < 0.0000001) {
			cerr << "\nnormalize() sees vector (0,0,0)!";
			return; // does nothing to zero vectors;
		}
		float scaleFactor = 1.0f/(float)sqrt(sizeSq);
		x *= scaleFactor; y *= scaleFactor; z *= scaleFactor;
	}
	Vector3(float xx, float yy, float zz) { x = xx; y = yy; z = zz; }
	Vector3(Vector3& v) { x = v.x; y = v.y; z = v.z; }
	Vector3() { x = y = z = 0; } //default constructor
	Vector3 cross(Vector3 b) { //return this cross b
	   Vector3 c(y*b.z - z*b.y, z*b.x - x*b.z, x*b.y - y*b.x);
	   return c;
	}
	float dot(Vector3 b) { // return this dotted with b
		return x * b.x + y * b.y + z * b.z;
	}
	Vector3 xfrmNormal(Affine4 a);
};

// @@@@@@@@@@@@@@@@@@@@@ Color3 class @@@@@@@@@@@@@@@@
class Color3 { // holds an red,green,blue 3-tuple
public:
	float red, green, blue;
	Color3() { red = green = blue = 0; }
	Color3(float r, float g, float b) { red = r; green = g; blue = b; }
	Color3(Color3& c) { red = c.red; green = c.green; blue = c.blue; }
	void set(float r, float g, float b) { red = r; green = g; blue = b; }
	void set(Color3& c) { red = c.red; green = c.green; blue = c.blue; }
	void add(float r, float g, float b) { red += r; green += g; blue += b; }
	void add(Color3& src, Color3& refl) { // add the product of source color and reflection coefficient
		red   += src.red   * refl.red;
		green += src.green * refl.green; 
		blue  += src.blue  * refl.blue;
	}
	void add(Color3& colr) { // add colr to this color
		red += colr.red ; green += colr.green; blue += colr.blue;
	}
	void build4tuple(float v[]) { // load 4-tuple with this color: v[3] = 1 for homogeneous
		v[0] = red; v[1] = green; v[2] = blue; v[3] = 1.0f;
	}
};

//@@@@@@@@@@@@@@@@@@@@ light class @@@@@@@@@@@@@@@@@@@
class Light { // for a linked list of light sources color and position
public:
	Point3 pos;
	Color3 color;
	Light* next;
	void setPosition(Point3 p) { pos.set(p); }
	void setColor(Color3 c) { color.set(c); }
	Light() { next = NULL; }
};
// @@@@@@@@@@@@@@@@@@@@@ Affine4 class @@@@@@@@@@@@@@@@
class Affine4 {// manages homogeneous affine transformations
// including inverse transformations
// and a stack to put them on
// used by Scene class to read SDL files
public:
	float m[16]; // hold a 4 by 4 matrix
	Affine4() { // make identity transform
		setIdentityMatrix();
	}
	//float det() {
	//	return (m[0]*m[5]*m[10]*m[15] + m[0]*m[6]*m[11]*m[13] + m[0]*m[7]*m[9]*m[14] +
	//		    m[1]*m[4]*m[11]*m[14] + m[1]*m[6]*m[8]*m[15] + m[1]*m[7]*m[10]*m[12] +
	//		    m[2]*m[4]*m[9]*m[15] + m[2]*m[5]*m[11]*m[12] + m[2]*m[7]*m[10]*m[13] +
	//		    m[3]*m[4]*m[10]*m[13] + m[3]*m[5]*m[8]*m[14] + m[3]*m[6]*m[9]*m[12] -
	//		    m[0]*m[5]*m[11]*m[14] - m[0]*m[6]*m[9]*m[15] - m[0]*m[7]*m[10]*m[13] -
	//		    m[1]*m[4]*m[10]*m[15] - m[1]*m[6]*m[11]*m[12] - m[1]*m[7]*m[8]*m[14] -
	//		    m[2]*m[4]*m[9]*m[13] - m[2]*m[5]*m[8]*m[15] - m[2]*m[7]*m[9]*m[12] -
	//		    m[3]*m[4]*m[9]*m[14] - m[3]*m[5]*m[10]*m[12] - m[3]*m[6]*m[8]*m[13]);
	//}
	void setIdentityMatrix() { // make identity transform
		m[0] = m[5]  = m[10] = m[15] = 1.0;
		m[1] = m[2]  = m[3]  = m[4]  = 0.0;
		m[6] = m[7]  = m[8]  = m[9]  = 0.0;
		m[11]= m[12] = m[13] = m[14] = 0.0;
	}
	void set(Affine4 a) { // set this matrix to a
		for(int i = 0; i < 16; i++)
			m[i]=a.m[i];
	}
	//<<<<<<<<<<<< postMult >>>>>>>>>>>
	void postMult(Affine4 n){ // postmultiplies this with n
		float sum;
		Affine4 tmp; 
		tmp.set(*this); // tmp copy
		for(int c = 0; c < 4; c++)// form this = tmp * n
			for(int r = 0; r <4 ; r++)
			{
				sum = 0;
				for(int k = 0; k < 4; k++)
					sum += tmp.m[4 * k + r]* n.m[4 * c + k];
				m[4 * c + r] = sum;
			} // end of for loops
	}
	//<<<<<<<<<<<< preMult >>>>>>>>>>>
	void preMult(Affine4 n){ // premultiplies this with n
		float sum;
		Affine4 tmp; 
		tmp.set(*this); // tmp copy
		for(int c = 0; c < 4; c++)// form this = tmp * n
			for(int r = 0; r <4 ; r++)
			{
				sum = 0;
				for(int k = 0; k < 4; k++)
					sum += n.m[4 * k + r]* tmp.m[4 * c + k];
				m[4 * c + r] = sum;
			} // end of for loops
	}
}; // end of Affine4 class

//@@@@@@@@@@ AffineNode class @@@@@@@@@@@
class AffineNode{
// used by Scene class to read SDL files
public:
	Affine4 * affn;
	Affine4 * invAffn;
	AffineNode * next;
	AffineNode() {
		next = NULL;
		affn = new Affine4; // new affine with identity in it
		invAffn = new Affine4; // and for the inverse
	}
	~AffineNode() { //destructor
		delete affn;
		delete invAffn;
	}
};

//@@@@@@@@@@@@@@@@ AffineStack class @@@@@@@@@@@@
class AffineStack{
// used by Scene class to read SDL files
public:
	AffineNode * tos;
	AffineStack() { // default constructor;puts identity on top
		tos = new AffineNode; // node with identity in it
		tos->next = NULL;
	}
	void dup() {
		AffineNode* tmp = new AffineNode;			
		tmp->affn = new Affine4(*(tos->affn));
		tmp->invAffn = new Affine4(*(tos->invAffn));
		tmp->next = tos;
		tos = tmp;
	}
	void setIdentity() { // make top item the identity matrix
		assert(tos != NULL);
		tos->affn->setIdentityMatrix();
		tos->invAffn->setIdentityMatrix();
	}
	void popAndDrop() {
		if(tos == NULL) return; // do nothing
		AffineNode *tmp = tos;
		tos = tos->next;
		delete tmp; // should call destructor, which deletes trices
	}	
	void releaseAffines() { // pop and drop all remaining items 
		while(tos) popAndDrop();
	}
	void rotate(float angle, Vector3 u) {
		Affine4 rm; // make identity matrix
		Affine4 invRm;
		u.normalize(); // make the rotation axis unit length
		float ang = angle * 3.14159265f/ 180.0f; // deg to  
		float c = cos(ang), s = sin(ang);
		float mc = 1.0f - c;
	 	//fill the 3x3 upper left matrix - Chap.5 p. 29
		rm.m[0] = c + mc * u.x * u.x;
		rm.m[1] = mc * u.x * u.y + s * u.z;
		rm.m[2] = mc * u.x * u.z - s * u.y;
		rm.m[4] = mc * u.y * u.x - s * u.z;
		rm.m[5] = c + mc * u.y * u.y;
		rm.m[6] = mc * u.y * u.z + s * u.x;
		rm.m[8] = mc * u.z * u.x + s * u.y;
		rm.m[9] = mc * u.z * u.y - s * u.x;
		rm.m[10] = c + mc * u.z * u.z;
		// same for inverse : just sign of s is changed
		invRm.m[0] = c + mc * u.x * u.x;
		invRm.m[1] = mc * u.x * u.y - s * u.z;
		invRm.m[2] = mc * u.x * u.z + s * u.y;
		invRm.m[4] = mc * u.y * u.x + s * u.z;
		invRm.m[5] = c + mc * u.y * u.y;
		invRm.m[6] = mc * u.y * u.z - s * u.x;
		invRm.m[8] = mc * u.z * u.x - s * u.y;
		invRm.m[9] = mc * u.z * u.y + s * u.x;
		invRm.m[10] = c + mc * u.z * u.z;
		tos->affn->postMult(rm);
		tos->invAffn->preMult(invRm);
	}
	void scale(float sx, float sy, float sz) { // post-multiply top item by scaling
		#define sEps 0.00001
		Affine4 scl;// make an identity
		Affine4 invScl;
		scl.m[0]  = sx; 
		scl.m[5]  = sy; 
		scl.m[10] = sz;// adjust it to a scaling matrix
		if(fabs(sx) < sEps || fabs(sy) < sEps || fabs(sz) < sEps) {
			cerr << "degenerate scaling transformation!\n";
		}
		invScl.m[0] = 1/sx;  invScl.m[5] = 1/sy;  invScl.m[10] = 1/sz;
		tos->affn->postMult(scl); //
		tos->invAffn->preMult(invScl);
	}		
	void translate(Vector3 d) {
		Affine4 tr; // make identity matrix
		Affine4 invTr;
		tr.m[3] = d.x;		tr.m[7] = d.y;		tr.m[11] = d.z;
		invTr.m[3] = -d.x;	invTr.m[7] = -d.y;	invTr.m[11] = -d.z;
		//tr.m[12] = d.x;		tr.m[13] = d.y;		tr.m[14] = d.z;
		//invTr.m[12] = -d.x;	invTr.m[13] = -d.y; invTr.m[14] = -d.z;
		tos->affn->postMult(tr);
		tos->invAffn->preMult(invTr);
	}
}; // end of AffineStack class